#!/usr/bin/env python3

"""Perform the metrics and video properties measures."""

import datetime
import json
import logging
import pathlib
import sqlite3

from context_verbose import Printer
from flufl.lock import Lock
import cutcutcodec
import numpy as np

from mendevi.database.serialize import list_to_binary
from mendevi.utils import compute_video_hash


def _fill_vid_table(database: pathlib.Path, vid_id: bytes, video: pathlib.Path, **kwargs):
    """Complete the table t_vid_video."""
    # get actual informations
    with sqlite3.connect(database) as sql_database:
        sql_database.row_factory = sqlite3.Row
        cursor = sql_database.cursor()
        prop = cursor.execute(
            "SELECT * FROM t_vid_video WHERE vid_id=?", (vid_id,)
        ).fetchone()
    prop = {} if prop is None else dict(prop)

    # missing fields
    fields = {
        "vid_codec",
        "vid_duration",
        "vid_eotf",
        "vid_fps",
        "vid_frames",
        "vid_gamut",
        "vid_height",
        "vid_name",
        "vid_pix_fmt",
        "vid_size",
        "vid_uvq",
        "vid_width",
    }
    fields -= {k for k, v in prop.items() if v is not None}
    if kwargs["uvq"] is False:
        fields -= {"vid_uvq"}
    if not fields:
        return
    fields = sorted(fields)  # to frozen order in sql request and for printing repetability

    # fill missing fields
    pad = max(map(len, fields)) - 4
    with Printer(f"Get properties of {video.name}...", color="green") as prt:
        for field in fields:
            match field:
                case "vid_duration":
                    prop["vid_duration"] = float(cutcutcodec.get_duration_video(video))
                    prt.print(f"{'duration':<{pad}}: ", end="")
                    prt.print_time(prop["vid_duration"], print_headers=False)
                case "vid_fps":
                    prop["vid_fps"] = float(cutcutcodec.get_rate_video(video))
                    prt.print(f"{'fps':<{pad}}: {prop['vid_fps']:.2f} Hz")
                case "vid_frames":
                    header, info = cutcutcodec.core.analysis.ffprobe.get_slices_metadata(video)
                    header, info = header[0], info[0]
                    frames = [dict(zip(header, l)) for l in info.tolist()]
                    prop["vid_frames"] = json.dumps(frames)
                    prt.print(f"{'frames':<{pad}}: {len(frames)} frames")
                case "vid_height":
                    prop["vid_height"] = cutcutcodec.get_resolution(video)[0]
                    prt.print(f"{'height':<{pad}}: {prop['vid_height']} pixels")
                case "vid_width":
                    prop["vid_width"] = cutcutcodec.get_resolution(video)[1]
                    prt.print(f"{'width':<{pad}}: {prop['vid_width']} pixels")
                case "vid_pix_fmt":
                    prop["vid_pix_fmt"] = cutcutcodec.get_pix_fmt(video)
                    prt.print(f"{'pix fmt':<{pad}}: {prop['vid_pix_fmt']}")
                case "vid_uvq":
                    uvq = cutcutcodec.compare(None, video, uvq=True)["uvq"]
                    prop["vid_uvq"] = list_to_binary(uvq)
                    prt.print(f"{'uvq':<{pad}}: {np.mean(uvq):.2f} +/- {np.std(uvq):.2f}")
                case "vid_size":
                    prop["vid_size"] = video.stat().st_size
                    prt.print(f"{'size':<{pad}}: {prop['vid_size']*1e-6:.2f} MB")
                case "vid_name":
                    prop["vid_name"] = video.name
                    prt.print(f"{'name':<{pad}}: {prop['vid_name']}")
                case "vid_eotf":
                    prop["vid_eotf"] = cutcutcodec.get_colorspace(video).transfer
                    prt.print(f"{'eotf':<{pad}}: {prop['vid_eotf']}")
                case "vid_gamut":
                    prop["vid_gamut"] = cutcutcodec.get_colorspace(video).primaries
                    prt.print(f"{'gamut':<{pad}}: {prop['vid_gamut']}")
                case "vid_codec":
                    prop["vid_codec"] = cutcutcodec.get_codec_video(video)
                    prt.print(f"{'codec':<{pad}}: {prop['vid_codec']}")

    # update result
    with (
        sqlite3.connect(database) as sql_database,
        Lock(str(database.with_name(".dblock")), lifetime=datetime.timedelta(seconds=60)),
    ):
        cursor = sql_database.cursor()
        cursor.execute(
            "INSERT OR IGNORE INTO t_vid_video (vid_id) VALUES (?)", (vid_id,)
        )
        cursor.execute(
            (
                f"UPDATE t_vid_video SET {', '.join(f'{k}=?' for k in fields)} "
                "WHERE vid_id=?"
            ),
            [prop[k] for k in fields] + [vid_id],
        )


def _fill_met_table(
    database: pathlib.Path,
    ref_id: bytes,
    ref_path: pathlib.Path,
    dis_id: bytes,
    dis_path: pathlib.Path,
    **kwargs,
):
    """Complete the table t_met_metric."""
    # get actual informations
    with sqlite3.connect(database) as sql_database:
        sql_database.row_factory = sqlite3.Row
        cursor = sql_database.cursor()
        metrics = cursor.execute(
            "SELECT * FROM t_met_metric WHERE met_ref_vid_id=? AND met_dis_vid_id=?",
            (ref_id, dis_id)
        ).fetchone()
    metrics = {} if metrics is None else dict(metrics)

    # missing fields
    fields = [
        *(("met_lpips_alex",) if kwargs["lpips_alex"] else ()),
        *(("met_lpips_vgg",) if kwargs["lpips_vgg"] else ()),
        *(("met_psnr",) if kwargs["psnr"] else ()),
        *(("met_ssim",) if kwargs["ssim"] else ()),
        *(("met_vmaf",) if kwargs["vmaf"] else ()),
    ]
    fields = set(fields)
    fields -= {k for k, v in metrics.items() if v is not None}
    if not fields:
        return
    fields = sorted(fields)  # to frozen order in sql request and for printing repetability

    # fill missing fields
    pad = max(map(len, fields)) - 4
    with Printer(
        f"Compute metrics between {ref_path.name} and {dis_path.name}...", color="green"
    ) as prt:
        new_metrics = cutcutcodec.compare(
            ref_path, dis_path,
            lpips_alex=kwargs["lpips_alex"],
            lpips_vgg=kwargs["lpips_vgg"],
            psnr=kwargs["psnr"],
            ssim=kwargs["ssim"],
            vmaf=kwargs["vmaf"],
        )
        for metric in sorted(new_metrics):
            values = new_metrics[metric]
            prt.print(f"{metric:<{pad}}: {np.mean(values):.2f} +/- {np.std(values):.2f}")
    metrics = {f"met_{m}": list_to_binary(v) for m, v in new_metrics.items()}

    # update result
    with (
        sqlite3.connect(database) as sql_database,
        Lock(str(database.with_name(".dblock")), lifetime=datetime.timedelta(seconds=60)),
    ):
        cursor = sql_database.cursor()
        cursor.execute(
            "INSERT OR IGNORE INTO t_met_metric (met_ref_vid_id, met_dis_vid_id) VALUES (?, ?)",
            (ref_id, dis_id)
        )
        cursor.execute(
            (
                f"UPDATE t_met_metric SET {', '.join(f'{k}=?' for k in fields)} "
                "WHERE met_ref_vid_id=? AND met_dis_vid_id=?"
            ),
            [metrics[k] for k in fields] + [ref_id, dis_id],
        )


def probe_and_store(database: pathlib.Path, video: pathlib.Path, **kwargs):
    """Measure the properties of the video.

    Parameters
    ----------
    database : pathlike
        The path of the existing database to be updated.
    video : pathlib.Path
        The source video file to be annalysed.
    **kwargs : dict
        All the metrics
    """
    assert isinstance(video, pathlib.Path), video.__class__.__name__
    assert video.is_file(), video

    vid_id: bytes = compute_video_hash(video)

    _fill_vid_table(database, vid_id, video, **kwargs)

    # get the references videos for comparative comparisons
    references: dict[pathlib.Path, bytes] = kwargs["ref"].copy()
    with sqlite3.connect(database) as sql_database:
        cursor = sql_database.cursor()
        for ref_name, ref_id in cursor.execute("""
            SELECT vid_name, enc_src_vid_id FROM t_enc_encode
            JOIN t_vid_video ON enc_src_vid_id=vid_id
            WHERE enc_dst_vid_id=?""",
            (vid_id,),
        ):
            # try to find video full path
            if (ref := database.with_name(ref_name)).is_file():
                references[ref] = ref_id
            elif (ref := video.with_name(ref_name)).is_file():
                references[ref] = ref_id
            else:
                logging.info("failed to find the reference video %s", ref)
    references = {ref: ref_id for ref, ref_id in references.items() if ref_id != vid_id}

    # perform the comparative comparisons
    for ref, ref_id in references.items():
        _fill_met_table(database, ref_id, ref, vid_id, video, **kwargs)



#!/usr/bin/env python3

"""Provide simple tools."""

import base64
import functools
import hashlib
import logging
import multiprocessing.pool
import numbers
import pathlib
import platform
import queue
import re
import threading
import time
import typing

from cutcutcodec.core.io import VIDEO_SUFFIXES
import cutcutcodec
import tqdm

from mendevi.g5kpower import g5kpower
from mendevi.psutil import Usage
from mendevi.rapl import RAPL

PATHLIKE = str | bytes | pathlib.Path


class Activity(threading.Thread):
    """Measure the computer activity of a section.

    Examples
    --------
    >>> import pprint, time
    >>> from mendevi.utils import Activity
    >>> with Activity() as activity:
    ...     time.sleep(1)
    ...
    >>> pprint.pprint(activity)
    >>>
    """

    def __init__(self, sleep: numbers.Real=50e-3):
        """Initalize the perf context.

        Parameters
        ----------
        sleep : float, default=50e-3
            The time interval between 2 measures (in s).
        """
        assert isinstance(sleep, numbers.Real), sleep.__class__.__name__
        assert sleep > 0, sleep

        super().__init__(daemon=True)

        self._rapl_catcher = RAPL(sleep=sleep, no_fail=True)
        self._usage_catcher = Usage(sleep=sleep)
        self._exit_queue = queue.Queue()
        self.sleep = float(sleep)
        self.res: dict = {}

    def run(self):
        """Perform the measures."""
        self.res["start"] = time.time()
        with self._rapl_catcher as rapl, self._usage_catcher as usage:
            self._exit_queue.get()  # wait
        self.res |= {
            "ps_core": usage["cpu"],
            "ps_cores": usage["cpus"],
            "ps_dt": usage["dt"],
            "ps_ram": usage["ram"],
        }
        if rapl is not None:
            self.res |= {
                "rapl_dt": rapl["dt"],
                "rapl_energy": rapl["energy"],
                "rapl_power": rapl["power"],
                "rapl_powers": rapl["powers"],
            }

    def __enter__(self) -> dict:
        r"""Start to measure.

        Returns
        -------
        activity: dict[str]
            * duration: float, the real measure duration.
            * ps_core: float, the mean cummulated usage of all the logical cpus.
            * ps_cores: list[list[float]], tensor of detailed usage of each logical core in %.
            * ps_dt: list[float], the duration of each interval (in s).
            * ps_ram: list[int], list of the sampled ram usage in bytes in each point.
            * rapl_dt: list[float], the duration of each interval (in s).
            * rapl_energy: float, the total energy consumption (in J).
            * rapl_power: float, the average power, energy divided by the duration (in w).
            * rapl_powers: list[float], the average power in watt in each interval.
            * start: float, absolute timestamp.
            * wattmeter_dt: list[float],  the duration of each interval (in s).
            * wattmeter_energy: float, the total energy consumption (in J).
            * wattmeter_power: float, the average power, energy divided by the duration (in w).
            * wattmeter_powers: list[float], the sampled power in watt in each point.

        Notes
        -----
        The returned dictionary is update inplace when we exit the code bloc.
        Only the successfull field are created.
        """
        self.start()
        return self.res

    def __exit__(self, *_):
        """Stop the measure and update the dictionary returnd by __enter__."""
        # stop
        self.res["duration"] = time.time() - self.res["start"]
        self._exit_queue.put(None)
        self.join()
        # request wattmeter power
        try:
            wattmeter = g5kpower(platform.node(), self.res["start"], self.res["duration"])
        except ValueError:
            wattmeter = None
        else:
            self.res |= {
                "wattmeter_dt": wattmeter["dt"],
                "wattmeter_energy": wattmeter["energy"],
                "wattmeter_power": wattmeter["power"],
                "wattmeter_powers": wattmeter["powers"],
            }


def compute_video_hash(
    videos: PATHLIKE | typing.Iterable[PATHLIKE]
) -> bytes | dict[pathlib.Path, bytes]:
    r"""Compute the checksum of the video.

    For :math:`n` hash of :math:`b` bits, the proba of the colision :math:`C` is
    :math:`p(C) = 1 - \left(\frac{2^k-1}{2^k}\right)^{\frac{n(n-1)}{2}}`.

    The md5 hash uses :math:`b = 128` bits. If we add one video per second durring 10 years,
    the proba of colision is about :math:`p(C) \approx 1.46*10^{-22}`.

    That's why the md5 hash is used to identify the video files.

    Parameters
    ----------
    videos : pathlike or list[pathlike]
        The single or set of video you want to compute the signature.

    Returns
    -------
    signatures
        The md5 checksum of the video file. In the case of a multiple file,
        a dictionary containing the file and the hash is returned rather a single hash.
        If the file does not exists, return None.
    """
    def _hash(video: PATHLIKE) -> pathlib.Path:
        video = pathlib.Path(video)
        if match := re.search(r"[2-7a-z]{26}", video.stem):
            return video, signature_to_hash(match.group())
        if not (video := video.expanduser()).is_file():
            return video, None
        with open(video, "rb") as raw:
            return video, hashlib.file_digest(raw, "md5").digest()

    if isinstance(videos, list | tuple | set | frozenset):
        with multiprocessing.pool.ThreadPool() as pool:
            return dict(tqdm.tqdm(
                pool.imap_unordered(_hash, videos),
                desc="compute videos checksum",
                dynamic_ncols=True,
                leave=False,
                smoothing=1e-4,
                total=len(videos),
                unit="video",
            ))
    return _hash(videos)[1]


@functools.cache
def get_pix_fmt(*args):
    """Alias to cutcutcodec func."""
    return cutcutcodec.get_pix_fmt(*args)


def get_project_root() -> pathlib.Path:
    """Return the absolute project root folder.

    Examples
    --------
    >>> from mendevi.utils import get_project_root
    >>> root = get_project_root()
    >>> root.is_dir()
    True
    >>> root.name
    'mendevi'
    >>> sorted(p.name for p in root.iterdir())  # doctest: +ELLIPSIS
    ['__init__.py', '__main__.py', ...]
    >>>
    """
    return pathlib.Path(__file__).resolve().parent


@functools.cache
def get_rate_video(*args):
    """Alias to cutcutcodec func."""
    return cutcutcodec.get_rate_video(*args)


@functools.cache
def get_resolution(*args):
    """Alias to cutcutcodec func."""
    return cutcutcodec.get_resolution(*args)


def hash_to_signature(checksum: bytes) -> str:
    r"""Convert the md5 binary hash value into an urlsafe string.

    Bijection of :py:func:`signature_to_hash`.

    Parameters
    ----------
    checksum : bytes
        The 128 bit binary hash value.

    Returns
    -------
    signature : str
        The 26 ascii [2-7a-z] symbols string of the converted checksum.

    Examples
    --------
    >>> from mendevi.utils import hash_to_signature
    >>> hash_to_signature(b"\xd4\x1d\x8c\xd9\x8f\x00\xb2\x04\xe9\x80\t\x98\xec\xf8B~")
    '2qoyzwmpaczaj2mabgmoz6ccpy'
    >>>
    """
    assert isinstance(checksum, bytes), checksum.__class__.__name__
    assert len(checksum) == 16, len(checksum)
    return base64.b32encode(checksum)[:26].decode().lower()


def signature_to_hash(signature: str) -> bytes:
    r"""Convert the string signature into the md5 checksum.

    Bijection of :py:func:`hash_to_signature`.

    Parameters
    ----------
    signature : str
        The 26 ascii [2-7a-z] symbols string of the converted checksum.

    Returns
    -------
    checksum : bytes
        The 128 bit binary hash value.

    Examples
    --------
    >>> from mendevi.utils import signature_to_hash
    >>> signature_to_hash("2qoyzwmpaczaj2mabgmoz6ccpy")
    b'\xd4\x1d\x8c\xd9\x8f\x00\xb2\x04\xe9\x80\t\x98\xec\xf8B~'
    >>>
    """
    assert isinstance(signature, str), signature.__class__.__name__
    assert re.fullmatch(r"[2-7a-z]{26}", signature), signature
    return base64.b32decode(f"{signature.upper()}======".encode())


def unfold_video_files(
    paths: typing.Iterable[PATHLIKE]
) -> typing.Iterable[pathlib.Path]:
    """Explore recursively the folders to find the video path.

    Parameters
    ----------
    paths : list[pathlike]
        All the folders, files, glob or recursive glob expression.

    Yields
    ------
    filename : pathlib.Path
        The path of the video.
    """
    assert hasattr(paths, "__iter__"), paths.__class__.__name__
    for path in paths:
        path = pathlib.Path(path).expanduser()
        if path.is_file():
            yield path
        elif path.is_dir():
            for root, _, files in path.walk():
                for file in files:
                    file = root / file
                    if file.suffix.lower() in VIDEO_SUFFIXES:
                        yield file
        elif "*" in path.name and path.parent.is_dir():
            yield from unfold_video_files(path.parent.glob(path.name))
        elif "**" in (parts := path.parts):
            idx = parts.index("**")
            yield from unfold_video_files(
                pathlib.Path(*parts[:idx]).glob(pathlib.Path(*parts[idx:]))
            )
        else:
            logging.warning("the path %s is not correct", path)



#!/usr/bin/env python3

"""Define the functions that enable values to be extracted from a select query."""

import functools
import numbers
import re

from mendevi.database.serialize import binary_to_list, binary_to_tensor


class SqlLinker:
    """Allow you to add an SQL query to an extractor."""

    def __init__(self, select: list[str], table: str, join: list[str]):
        """Initialise the linker.

        Parameters
        ----------
        select : list[str]
            The field to be returned (juste after SELECT), with the optional alias.
        table : str
            The data recovery table name, define after the FROM keyword.
        join : list[str]
            The joins, defined after the JOIN keyword.
        """
        # clean inputs
        self.select = [re.sub(r"\s+", " ", s) for s in select]
        self.table = re.sub(r"\s+", " ", table)
        self.join = [re.sub(r"\s+", " ", j) for j in join]

    @property
    def sql(self) -> str:
        """Write the sql request."""
        select_str = f"SELECT {', '.join(self.select)}"
        if len(select_str) >= 80:
            select_str = f"SELECT\n    {',\n    '.join(self.select)}"
        table_str = f"FROM {self.table}"
        if (join_str := "\n".join(re.sub(" ON ", "\n    ON ", j) for j in self.join)):
            sql = f"{select_str}\n{table_str}\n{join_str}"
        else:
            sql = f"{select_str}\n{table_str}"
        return sql

    def __call__(self, func: callable) -> callable:
        """Decorate a function.

        Returns
        -------
        A decorated function with the select, table, and join parameters.
        The docstring of the decorated function is also modified
        to illustrate the minimal SQL query with an example.
        """
        # set attributes
        func.select = self.select
        func.table = self.table
        func.join = self.join

        # set doctrsing
        doc: list[str] = (func.__doc__ or "").split("\n")
        example = (
            "\n"
            ".. code:: sql\n"
            "\n"
            f"    {'\n    '.join(self.sql.split('\n'))}"
            "\n"
        )
        doc.insert(1, example)
        func.__doc__ = "\n".join(doc)

        return func


@SqlLinker(
    ["enc_env_id", "enc_cmd", "enc_vid.vid_name AS enc_src_vid_name"],
    "t_enc_encode",
    ["JOIN t_vid_video AS enc_vid ON t_enc_encode.enc_src_vid_id = enc_vid.vid_id"],
)
def extract_enc_scenario(raw: dict[str]) -> str:
    """Return the unique string specific to the encoding scenario.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    Returns
    -------
    scenario : str
        Unique string specific to the encoding scenario.
    """
    assert isinstance(raw, dict), raw.__class__.__name__
    assert "enc_env_id" in raw, "Please correct the SQL query."
    assert "enc_cmd" in raw, "Please correct the SQL query."
    assert "enc_src_vid_name" in raw, "Please correct the SQL query."
    env_id, cmd, vid_name = raw["enc_env_id"], raw["enc_cmd"], raw["enc_src_vid_name"]
    assert isinstance(env_id, numbers.Integral), env_id.__class__.__name__
    assert isinstance(cmd, str), cmd.__class__.__name__
    assert isinstance(vid_name, str), vid_name.__class__.__name__
    return f"env {env_id}: {cmd.replace('src.mp4', vid_name)}"


@SqlLinker(
    ["enc_act.act_ps_dt AS enc_act_ps_dt", "enc_act.act_ps_core AS enc_act_ps_core"],
    "t_enc_encode",
    ["JOIN t_act_activity AS enc_act ON t_enc_encode.enc_act_id = enc_act.act_id"],
)
def extract_enc_cores(raw: dict[str]) -> float:
    """Return the average cumulative utilisation rate of logic cores during encoding.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    Returns
    -------
    enc_cores : int
        Average cumulative utilisation rate of logic cores during encoding.
    """
    assert isinstance(raw, dict), raw.__class__.__name__
    assert "enc_act_ps_dt" in raw, "Please correct the SQL query."
    assert "enc_act_ps_core" in raw, "Please correct the SQL query."
    act_ps_dt = binary_to_list(raw["enc_act_ps_dt"])
    act_ps_core = binary_to_tensor(raw["enc_act_ps_core"]).sum(axis=1)
    integral = (act_ps_core * act_ps_dt).sum()  # act_ps_core is already the average on each dt
    average = integral / act_ps_dt.sum()
    return float(average) / 100.0  # normalisation


@SqlLinker(
    ["enc_act.act_duration AS enc_act_duration"],
    "t_enc_encode",
    ["JOIN t_act_activity AS enc_act ON t_enc_encode.enc_act_id = enc_act.act_id"],
)
def extract_enc_duration(raw: dict[str]) -> float:
    """Return the video encoding time in seconds.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    Returns
    -------
    enc_duration : float
        Video encoding time in seconds.
    """
    assert isinstance(raw, dict), raw.__class__.__name__
    assert "enc_act_duration" in raw, "Please correct the SQL query."
    enc_duration = raw["enc_act_duration"]
    assert isinstance(enc_duration, numbers.Real), enc_duration.__class__.__name__
    assert enc_duration > 0.0, enc_duration.__class__.__name__
    return float(enc_duration)


@SqlLinker(["enc_effort"], "t_enc_encode", [])
def extract_enc_effort(raw: dict[str]) -> str:
    """Return the effort provided as a parameter to the encoder.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    Returns
    -------
    enc_effort : str
        Effort provided as a parameter to the encoder.
    """
    assert isinstance(raw, dict), raw.__class__.__name__
    assert "enc_effort" in raw, "Please correct the SQL query."
    enc_effort = raw["enc_effort"]
    assert isinstance(enc_effort, str), enc_effort.__class__.__name__
    return str(enc_effort)


@SqlLinker(["enc_threads"], "t_enc_encode", [])
def extract_enc_threads(raw: dict[str]) -> int:
    """Return the number of threads provided as a parameter to the encoder.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    Returns
    -------
    enc_threads : int
        Number of threads provided as a parameter to the encoder.
    """
    assert isinstance(raw, dict), raw.__class__.__name__
    assert "enc_threads" in raw, "Please correct the SQL query."
    enc_threads = raw["enc_threads"]
    assert isinstance(enc_threads, numbers.Integral), enc_threads.__class__.__name__
    assert enc_threads >= 1, enc_threads.__class__.__name__
    return int(enc_threads)


@SqlLinker(["enc_quality"], "t_enc_encode", [])
def extract_enc_quality(raw: dict[str]) -> float:
    """Return the quality level passed to the encoder.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    Returns
    -------
    quality : float
        Quality level passed to the encoder.
    """
    assert isinstance(raw, dict), raw.__class__.__name__
    assert "enc_quality" in raw, "Please correct the SQL query."
    enc_quality = raw["enc_quality"]
    assert isinstance(enc_quality, numbers.Real), enc_quality.__class__.__name__
    assert 0.0 <= enc_quality <= 1.0, enc_quality
    return float(enc_quality)


@SqlLinker(["enc_encoder"], "t_enc_encode", [])
def extract_enc_encoder(raw: dict[str]) -> str:
    """Return the name of the encoder.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    Returns
    -------
    encoder : str
        Name of the encoder.
    """
    assert isinstance(raw, dict), raw.__class__.__name__
    assert "enc_encoder" in raw, "Please correct the SQL query."
    enc_encoder = raw["enc_encoder"]
    assert isinstance(enc_encoder, str), enc_encoder.__class__.__name__
    return str(enc_encoder)


@SqlLinker(
    [
        "enc_act.act_wattmeter_dt AS enc_act_act_wattmeter_dt",
        "enc_act.act_wattmeter_power AS enc_act_wattmeter_power",
    ],
    "t_enc_encode",
    ["JOIN t_act_activity AS enc_act ON t_enc_encode.enc_act_id = enc_act.act_id"],
)
def extract_enc_wattmeter_energy(raw: dict[str]) -> float:
    """Return the total encoding energy consumption in Joules.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    Returns
    -------
    enc_power : float
        Total encoding energy consumption in Joules.
    """
    assert isinstance(raw, dict), raw.__class__.__name__
    assert "enc_act_act_wattmeter_dt" in raw, "Please correct the SQL query."
    assert "enc_act_wattmeter_power" in raw, "Please correct the SQL query."
    act_dt = binary_to_list(raw["enc_act_act_wattmeter_dt"])
    act_power = binary_to_list(raw["enc_act_wattmeter_power"])
    integral = 0.5 * ((act_power[:-1] + act_power[1:]) * act_dt).sum()  # trapez method
    return float(integral)


@SqlLinker(
    [
        "enc_act.act_wattmeter_dt AS enc_act_act_wattmeter_dt",
        "enc_act.act_wattmeter_power AS enc_act_wattmeter_power",
    ],
    "t_enc_encode",
    ["JOIN t_act_activity AS enc_act ON t_enc_encode.enc_act_id = enc_act.act_id"],
)
def extract_enc_wattmeter_power(raw: dict[str]) -> float:
    """Return the average encoding power consumption in Watts.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    Returns
    -------
    enc_power : float
        Average encoding power consumption in Watts.
    """
    assert isinstance(raw, dict), raw.__class__.__name__
    assert "enc_act_act_wattmeter_dt" in raw, "Please correct the SQL query."
    assert "enc_act_wattmeter_power" in raw, "Please correct the SQL query."
    act_dt = binary_to_list(raw["enc_act_act_wattmeter_dt"])
    act_power = binary_to_list(raw["enc_act_wattmeter_power"])
    integral = 0.5 * ((act_power[:-1] + act_power[1:]) * act_dt).sum()  # trapez method
    return float(integral / act_dt.sum())


@SqlLinker(["enc_width"], "t_enc_encode", [])
def extract_profile(raw: dict[str]) -> str:
    """Return the profile of the encoded video.

    The profile is determined based on the width of the video.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    Returns
    -------
    profile : str
        Profile of the encoded video.
    """
    from mendevi.cst.profiles import PROFILES
    assert isinstance(raw, dict), raw.__class__.__name__
    assert "enc_width" in raw, "Please correct the SQL query."
    enc_width = raw["enc_width"]
    assert isinstance(enc_width, numbers.Integral), enc_width.__class__.__name__
    dist_to_profile = {abs(v["resolution"][1]-enc_width): p for p, v in PROFILES.items()}
    return dist_to_profile[min(dist_to_profile)]


@SqlLinker(
    ["enc_vid.vid_name AS enc_src_vid_name"],
    "t_enc_encode",
    ["JOIN t_vid_video AS enc_vid ON t_enc_encode.enc_src_vid_id = enc_vid.vid_id"],
)
def extract_src_vid(raw: dict[str]) -> str:
    """Return the input video name.

    Parameters
    ----------
    raw : dict[str]
        The result line of select request.

    Returns
    -------
    vid_name : str
        Input video name.
    """
    from mendevi.cst.profiles import PROFILES
    assert isinstance(raw, dict), raw.__class__.__name__
    assert "enc_src_vid_name" in raw, "Please correct the SQL query."
    enc_src_vid_name = raw["enc_src_vid_name"]
    assert isinstance(enc_src_vid_name, str), enc_src_vid_name.__class__.__name__
    vid_name = re.sub(r"^reference_(\w+)_(?:sd|hd|fhd|uhd4k)\.\w+$", r"\1", enc_src_vid_name)
    return str(vid_name)



#!/usr/bin/env python3

"""CLI entry point to visualize database data."""

import ast
import importlib
import pathlib

from context_verbose import Printer
from pygments import highlight
from pygments.formatters import TerminalFormatter
from pygments.lexers import PythonLexer
import click

from mendevi.plot.axis import NAMES, get_label_extractor
from mendevi.plot.printer import printer
from .parse import parse_videos_database

NAMES_DOC = "        ".join(
    f"* {n}: {get_label_extractor(n)[0]}.\n" for n in NAMES
)
DOCSTRING = f"""Draw a chart from a database.

\b
Parameters
----------
database : pathlike, optional
    The path to the database where all measurements are stored.
    If a folder is provided, the database is created inside this folder.
x, y : tuple[str]
    The name of the x and y axes on each window. Possible values are:
        {NAMES_DOC}
\b
color, marker : str, optional
    All points with the same value for this variable will be displayed in the same colour and marker.
    Conversely, this allows points that do not share this criterion to be visually separated.
    The possible values are the same as for the x and y parameters.
window : str, optional
    Rather than putting everything in the same chart, this allows you to create sub-windows.
    There will be as many sub-windows as there are different values taken by this criterion.
    The possible values are the same as for the x and y parameters.
window_x, window_y : str, optional
    Same as window, but with precision on the axis on which to develop.
filter : str, optional
    This allows you to filter the data to select only a portion of it.
    The character string provided must be a Boolean Python expression
    that returns True to keep the data. For example: ``enc_duration > 10 and profile == "fhd"``.
"""


def _parse_args(prt: Printer, kwargs: dict):
    """Verification of the arguments."""
    # x
    assert "x" in kwargs, sorted(kwargs)
    assert isinstance(kwargs["x"], tuple), kwargs["x"].__class__.__name__
    assert kwargs["x"], "you must provide the -x option"
    assert all(isinstance(x, str) and x in NAMES for x in kwargs["x"]), kwargs["x"]
    prt.print(f"x         : {', '.join(kwargs['x'])}")

    # y
    assert "y" in kwargs, sorted(kwargs)
    assert isinstance(kwargs["y"], tuple), kwargs["y"].__class__.__name__
    assert kwargs["y"], "you must provide the -y option"
    assert all(isinstance(y, str) and y in NAMES for y in kwargs["y"]), kwargs["y"]
    prt.print(f"y         : {', '.join(kwargs['y'])}")

    # color
    kwargs["color"] = kwargs.get("color", None)
    if kwargs["color"] is not None:
        assert isinstance(kwargs["color"], str), kwargs["color"].__class__.__name__
        assert kwargs["color"] in NAMES, kwargs["color"]
        prt.print(f"color     : {kwargs['color']}")

    # marker
    kwargs["marker"] = kwargs.get("marker", None)
    if kwargs["marker"] is not None:
        assert isinstance(kwargs["marker"], str), kwargs["marker"].__class__.__name__
        assert kwargs["marker"] in NAMES, kwargs["marker"]
        prt.print(f"marker    : {kwargs['marker']}")

    # window
    kwargs["window"] = kwargs.get("window", None)
    if kwargs["window"] is not None:
        match (  # preference for y axis
            kwargs.get("window_x", None) is None,
            kwargs.get("window_y", None) is None,
            len(kwargs["x"]) == 1,
            len(kwargs["y"]) == 1,
        ):
            case (False, True, False, True):
                kwargs["window_y"] = kwargs["window"]
            case (True, False, True, False):
                kwargs["window_x"] = kwargs["window"]
            case (True, True, False, True):
                kwargs["window_y"] = kwargs["window"]
            case (True, True, True, False):
                kwargs["window_x"] = kwargs["window"]
            case (False, True, False, False):
                kwargs["window_y"] = kwargs["window"]
            case (True, False, False, False):
                kwargs["window_x"] = kwargs["window"]
            case _:
                kwargs["window_y"] = kwargs["window"]

    # window_x
    kwargs["window_x"] = kwargs.get("window_x", None)
    if kwargs["window_x"] is not None:
        assert isinstance(kwargs["window_x"], str), kwargs["window_x"].__class__.__name__
        assert kwargs["window_x"] in NAMES, kwargs["window_x"]
        prt.print(f"window x  : {kwargs['window_x']}")

    # window_y
    kwargs["window_y"] = kwargs.get("window_y", None)
    if kwargs["window_y"] is not None:
        assert isinstance(kwargs["window_y"], str), kwargs["window_y"].__class__.__name__
        assert kwargs["window_y"] in NAMES, kwargs["window_y"]
        prt.print(f"window y  : {kwargs['window_y']}")

    # filter
    kwargs["filter"] = kwargs.get("filter", None)
    if kwargs["filter"] is not None:
        assert isinstance(kwargs["filter"], str), kwargs["filter"].__class__.__name__
        try:
            expr = ast.parse(kwargs["filter"], mode="eval")
        except SyntaxError as err:
            raise SyntaxError(
                f"the option --filter {kwargs['filter']!r} is not python valid"
            ) from err
        names = {n.id for n in ast.walk(expr) if isinstance(n, ast.Name)}
        assert names, f"the filter {kwargs['filter']!r} has no effect"
        assert names.issubset(set(NAMES)), \
            f"{names-set(NAMES)} are invalid names, it has to be one of {NAMES}"
        prt.print(f"filter    : {kwargs['filter']}")


@click.command(help=DOCSTRING)
@click.argument("database", type=click.Path())
@click.option(
    "-x",
    type=click.Choice(NAMES),
    multiple=True,
    help="The code for the value to be displayed on the x-axis.",
)
@click.option(
    "-y",
    type=click.Choice(NAMES),
    multiple=True,
    help="The code for the value to be displayed on the y-axis.",
)
@click.option(
    "-c", "--color",
    type=click.Choice(NAMES),
    help="The discriminating criterion for colour.",
)
@click.option(
    "-m", "--marker",
    type=click.Choice(NAMES),
    help="The discriminating criterion for marker.",
)
@click.option(
    "-w", "--window",
    type=click.Choice(NAMES),
    help="The criterion on which to separate into sub-windows."
)
@click.option(
    "--window-x",
    type=click.Choice(NAMES),
    help="The criterion on which to separate into sub-windows along x axis."
)
@click.option(
    "--window-y",
    type=click.Choice(NAMES),
    help="The criterion on which to separate into sub-windows along y axis."
)
@click.option(
    "--filter", "-f",
    type=str,
    help="Data selection python expression",
)
def main(database: str, **kwargs):
    # parse args
    with Printer("Parse configuration...") as prt:
        _, database = parse_videos_database(prt, (), database)
        _parse_args(prt, kwargs)

    # get title name
    path = (
        pathlib.Path.cwd()
        / f"{'_'.join(kwargs['x'])}_as_a_function_of_{'_'.join(kwargs['y'])}.py"
    )

    # get code content
    code = printer(database=database, **kwargs)
    print("**********************************PYTHON CODE**********************************")
    print(highlight(code, PythonLexer(), TerminalFormatter()))
    print("*******************************************************************************")
    with open(path, "w", encoding="utf-8") as file:
        file.write(code)

    # excecute code
    spec = importlib.util.spec_from_file_location(path.stem, path)
    modulevar = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(modulevar)
    modulevar.main(modulevar.read_sql(database))

# fill the documentation
main.__doc__ = DOCSTRING



#!/usr/bin/env python3

"""Perform encoding measures."""

import datetime
import hashlib
import pathlib
import re
import shlex
import shutil
import sqlite3
import subprocess
import tempfile

from context_verbose import Printer
from flufl.lock import Lock
import cutcutcodec
import numpy as np
import tqdm

from mendevi.convert import get_convert_cmd
from mendevi.database.serialize import list_to_binary, tensor_to_binary
from mendevi.utils import compute_video_hash, hash_to_signature
from .utils import Activity


ENCODERS = {"libx264", "libx265", "libvpx-vp9", "libsvtav1", "vvc"}


def encode(src: pathlib.Path, **kwargs) -> tuple[pathlib.Path, str, dict[str]]:
    """Transcode an existing video.

    Parameters
    ----------
    src : pathlib.Path
        The source video file to be transcoded.
    **kwargs : dict
        Transmitted to :py:func:`get_transcode_cmd`.

    Returns
    -------
    dst : pathlib.Path
        The transcoded video path. The stem contains the md5 hash of the file content.
    cmd : str
        The ffmpeg command.
    activity : dict[str]
        The computeur activity during the transcoding process.
    """
    assert isinstance(src, pathlib.Path), src.__class__.__name__
    assert src.is_file(), src

    # find tempfile name
    signature = hashlib.md5(
        bytes(src) + " ".join(str(kwargs[k]) for k in sorted(kwargs)).encode("utf-8")
    ).hexdigest()
    dst = pathlib.Path(tempfile.gettempdir()) / f"{signature}.mp4"

    # get cmd
    cmd = get_transcode_cmd(src, dst, **kwargs)

    # display
    prt_cmd = " ".join(
        map(shlex.quote, [{str(src): "src.mp4", str(dst): "dst.mp4"}.get(c, c) for c in cmd])
    )
    with Printer(prt_cmd, color="green") as prt:
        prt.print(f"input video: {src.name}")
        load = tqdm.tqdm(
            dynamic_ncols=True,
            leave=False,
            smoothing=1e-4,
            total=round(float(cutcutcodec.get_duration_video(src)), 2),
            unit="s",
        )

        # transcode
        with Activity() as activity, subprocess.Popen(cmd, stderr=subprocess.PIPE) as process:
            signature = b""
            is_finish = False
            while not is_finish:
                while (
                    match := re.search(
                        br"time=(?P<h>\d+):(?P<m>\d{1,2}):(?P<s>\d{1,2}\.\d*)", signature
                    )
                ) is None:
                    if not (buff := process.stderr.read(32)):
                        is_finish = True
                        break
                    signature += buff
                else:
                    signature = signature[match.endpos:]
                    elapsed = round(
                        3600.0*float(match["h"]) + 60.0*float(match["m"]) + float(match["s"]),
                        2,
                    )
                    load.total = max(load.total, elapsed)
                    load.update(elapsed-load.n)
            load.close()
            if process.returncode:
                raise RuntimeError(f"failed to execute {cmd}")

        # print
        prt.print(f"avg cpu usage: {activity['ps_core']:.1f} %")
        prt.print(f"avg ram usage: {1e-9*np.mean(activity['ps_ram']):.2g} Go")
        if "rapl_power" in activity:
            prt.print(f"avg rapl power: {activity['rapl_power']:.2g} W")
        if "wattmeter_power" in activity:
            prt.print(f"avg wattmeter power: {activity['wattmeter_power']:.2g} W")

        # compute file hash
        signature = hash_to_signature(compute_video_hash(dst))
        prt.print(f"output video: sample_{signature}.mp4")

    # move file
    final_dst = src.parent / f"sample_{signature}.mp4"
    if not final_dst.exists():
        shutil.copy(dst, src.parent / f"sample_{signature}_partial.mp4")
        shutil.move(src.parent / f"sample_{signature}_partial.mp4", final_dst)
        final_dst.chmod(0o777)
    dst.unlink()

    return final_dst, prt_cmd, activity


def encode_and_store(
    database: pathlib.Path,
    env_id: int,
    src: pathlib.Path,
    **kwargs,
):
    """Transcode a video file and store the result in the database.

    Parameters
    ----------
    database : pathlike
        The path of the existing database to be updated.
    **kwargs
        Transmitted to :py:func:`encode`.

    Examples
    --------
    >>> import pathlib, tempfile
    >>> from mendevi.database.complete import add_environement
    >>> from mendevi.database.create import create_database
    >>> from mendevi.encode import encode_and_store
    >>> src = pathlib.Path("/data/dataset/video/despacito.mp4")
    >>> create_database(database := pathlib.Path(tempfile.mktemp(suffix=".sqlite")))
    >>> env_id = add_environement(database)
    >>> encode_and_store(
    ...     database, env_id, src,
    ...     encoder="libx264", profile="sd", effort="fast", quality=0.5, threads=8
    ... )
    >>> database.unlink()
    >>>
    """
    # transcode the video
    dst, cmd, activity = encode(src, **kwargs)

    with (
        sqlite3.connect(database) as sql_database,
        Lock(str(database.with_name(".dblock")), lifetime=datetime.timedelta(seconds=60)),
    ):
        cursor = sql_database.cursor()

        # fill video table
        try:
            cursor.execute(
                "INSERT INTO t_vid_video (vid_id, vid_name) VALUES (?, ?)",
                (kwargs["src_vid_id"], src.name)
            )
        except sqlite3.IntegrityError:
            pass
        dst_vid_id: bytes = compute_video_hash(dst)
        try:
            cursor.execute(
                "INSERT INTO t_vid_video (vid_id, vid_name) VALUES (?, ?)",
                (dst_vid_id, dst.name)
            )
        except sqlite3.IntegrityError:
            pass

        # fill activity table
        activity = {
            "act_duration": activity["duration"],
            "act_ps_core": tensor_to_binary(activity["ps_cores"]),
            "act_ps_dt": list_to_binary(activity["ps_dt"]),
            "act_ps_ram": list_to_binary(activity["ps_ram"]),
            "act_rapl_dt": list_to_binary(activity.get("rapl_dt", None)),
            "act_rapl_power": list_to_binary(activity.get("rapl_powers", None)),
            "act_start": activity["start"],
            "act_wattmeter_dt": list_to_binary(activity.get("wattmeter_dt", None)),
            "act_wattmeter_power": list_to_binary(activity.get("wattmeter_powers", None)),
        }
        keys = list(activity)
        (act_id,) = cursor.execute(
            (
                f"INSERT INTO t_act_activity ({', '.join(keys)}) "
                f"VALUES ({', '.join('?'*len(keys))}) RETURNING act_id"
            ),
            [activity[k] for k in keys],
        ).fetchone()

        # fill encode table
        values = {
            "enc_act_id": act_id,
            "enc_cmd": cmd,
            "enc_dst_vid_id": dst_vid_id,
            "enc_effort": kwargs["effort"],
            "enc_encoder": kwargs["encoder"],
            "enc_env_id": env_id,
            "enc_fps": float(kwargs["fps"]),
            "enc_height": kwargs["resolution"][0],
            "enc_pix_fmt": kwargs["pix_fmt"],
            "enc_quality": kwargs["quality"],
            "enc_src_vid_id": kwargs["src_vid_id"],
            "enc_threads": kwargs["threads"],
            "enc_width": kwargs["resolution"][1],
        }
        keys = list(values)
        cursor.execute(
            f"INSERT INTO t_enc_encode ({', '.join(keys)}) VALUES ({', '.join('?'*len(keys))})",
            [values[k] for k in keys]
        )


def get_transcode_cmd(src: pathlib.Path, dst: pathlib.Path, **kwargs) -> list[str]:
    """Return the ffmpeg encode cmd."""

    def libsvtav1_lp(threads: int) -> int:
        """Convert threads in parralel level."""
        # https://gitlab.com/AOMediaCodec/SVT-AV1/-/blob/master/Source/Lib/Globals/enc_handle.c#L598
        # lp=1 -> threads=1
        # lp=2 -> threads=2
        # lp=3 -> threads=8
        # lp=4 -> threads=12
        # lp=5 -> threads=16
        # lp=6 -> threads=20
        return {  # threads to lp
            1: 1,
            2: 2, 3: 2, 4: 2, 5: 2,
            6: 3, 7: 3, 8: 3, 9: 3, 10: 3,
            11: 4, 12: 4, 13: 4, 14: 4,
            15: 5, 16: 5, 17: 5,
        }.get(threads, 6)

    # header
    cmd: list[str] = ["ffmpeg", "-y", "-i", str(src)]

    # filter
    if (
        filter_cmd := get_convert_cmd(
            src,
            additional_filter=kwargs["filter"],
            fps=kwargs["fps"],
            pix_fmt=kwargs["pix_fmt"],
            resolution=kwargs["resolution"],
        )
    ):
        cmd.extend(["-vf", filter_cmd])

    # transcode
    cmd.append("-c:v")
    match kwargs["encoder"]:
        case "libx264":
            cmd.extend([  # https://ffmpeg.party/x264/
                "libx264",
                "-crf", str(round(kwargs["quality"]*51.0, 1)),
                "-preset", kwargs["effort"],
                "-tune", "ssim",
                "-threads", str(kwargs["threads"]),
            ])
        case "libx265":
            cmd.extend([   # https://x265.readthedocs.io/en/master/cli.html
                "libx265",
                "-crf", str(round(kwargs["quality"]*51.0, 1)),
                "-preset", kwargs["effort"],
                "-tune", "ssim",
                "-x265-params",
                (
                    f"frame-threads={kwargs['threads']}:"
                    f"pools={kwargs['threads']}:"
                    f"wpp={1 if kwargs['threads'] != 1 else 0}"
                ),
            ])
        case "libvpx-vp9":
            cmd.extend([  # https://wiki.webmproject.org/ffmpeg/vp9-encoding-guide
                "libvpx-vp9",
                "-crf", str(round(kwargs["quality"]*63.0)),
                # in [-16, 16]
                "-speed", {"slow": "-2", "medium": "1", "fast": "8"}[kwargs["effort"]],
                "-tune", "ssim",
                "-row-mt", "1", "-threads", str(kwargs["threads"]),
            ])
        case "libsvtav1":
            cmd.extend([
                "libsvtav1",
                "-crf", str(round(kwargs["quality"]*63.0)),
                "-preset", {"slow": "4", "medium": "6", "fast": "8"}[kwargs["effort"]],
                # tune=2 broken in v3.0.2
                "-svtav1-params", f"film-grain=0:lp={libsvtav1_lp(kwargs['threads'])}",
            ])
        case "vvc":
            bit = int(re.search(r"(?P<bit>\d+)le", kwargs["pix_fmt"] + "8le")["bit"])
            cmd.extend([
                "vvc",
                "-qp", str(round(kwargs["quality"]*63.0)),
                "-preset", kwargs["effort"],
                "-qpa", "1",
                "-vvenc-params", f"internalbitdepth={bit}",
                "-threads", str(kwargs['threads']),
            ])

    # final
    cmd.append(str(dst))
    return cmd



#!/usr/bin/env python3

"""Extract the values associated with an axis."""

import numbers

from . import extract


NAMES = [
    "enc_cores",
    "enc_duration",
    "enc_effort", "effort", "preset",
    "enc_encoder", "encoder",
    "enc_quality", "quality",
    "enc_scenario",
    "enc_threads", "threads",
    "enc_wattmeter_energy", "enc_energy",
    "enc_wattmeter_power", "enc_power",
    "profile",
    "src_vid",
]


def get_label_extractor(name: str):
    """Get the way to deserialize a raw value.

    Parameters
    ----------
    name : str
        The value code, one of :py:cst`mendevi.plot.axis.NAMES`.

    Returns
    -------
    label : str
        The description of the physical quantity.
        This description can be used to label the axes of a graph.
    func : callable
        The function that performs the verification and deserialisation task.
    """
    assert isinstance(name, str), name.__class__.__name__
    match name:
        case "enc_cores":
            return (
                "Average cumulative utilisation rate of logic cores during encoding",
                extract.extract_enc_cores,
            )
        case "enc_effort" | "effort" | "preset":
            return (
                "Effort provided as a parameter to the encoder",
                extract.extract_enc_effort,
            )
        case "enc_duration":
            return (
                "Video encoding time in seconds",
                extract.extract_enc_duration,
            )
        case "enc_encoder" | "encoder":
            return (
                "Name of the encoder",
                extract.extract_enc_encoder,
            )
        case "enc_quality" | "quality":
            return (
                "Quality level passed to the encoder",
                extract.extract_enc_quality,
            )
        case "enc_scenario":
            return (
                "Unique string specific to the encoding scenario",
                extract.extract_enc_scenario,
            )
        case "enc_threads" | "threads":
            return (
                "Number of threads provided as a parameter to the encoder",
                extract.extract_enc_threads,
            )
        case "enc_wattmeter_energy" | "enc_energy":
            return (
                "Total encoding energy consumption in Joules",
                extract.extract_enc_wattmeter_energy,
            )
        case "enc_wattmeter_power" | "enc_power":
            return (
                "Average encoding power consumption in Watts",
                extract.extract_enc_wattmeter_power,
            )
        case "profile":
            return (
                "Profile of the encoded video",
                extract.extract_profile,
            )
        case "src_vid":
            return (
                "Input video name",
                extract.extract_src_vid,
            )



#!/usr/bin/env python3

"""Perform decoding measures."""

import itertools
import pathlib
import sqlite3

from context_verbose import Printer
import click

from mendevi.database.complete import add_environement
from mendevi.decode import decode_and_store
from mendevi.utils import compute_video_hash, get_resolution
from .parse import ResolutionParamType, parse_videos_database



def _parse_args(prt: Printer, kwargs: dict):
    """Verification of the arguments."""
    # repeat
    assert "repeat" in kwargs, sorted(kwargs)
    assert isinstance(kwargs["repeat"], int), kwargs["repeat"].__class__.__name__
    assert kwargs["repeat"] >= 1, kwargs["repeat"]
    prt.print(f"repeat    : {kwargs['repeat']}")

    # resolution
    assert "resolution" in kwargs, sorted(kwargs)
    assert isinstance(kwargs["resolution"], tuple), kwargs["resolution"].__class__.__name__
    assert all(
        isinstance(r, tuple) and len(r) == 2 and all(isinstance(s, int) and s > 0 for s in r)
        for r in kwargs["resolution"]
    ), kwargs["resolution"]
    if kwargs["resolution"]:
        prt.print(f"resolution: {', '.join(f'w={w}:h={h}' for h, w in kwargs['resolution'])}")
    else:
        kwargs["resolution"] = (None,)

        # filter
    kwargs["filter"] = kwargs.get("filter", ())
    assert isinstance(kwargs["filter"], tuple), kwargs["filter"].__class__.__name__
    assert all(isinstance(f, str) for f in kwargs["filter"]), kwargs["filter"]
    if kwargs["filter"]:
        prt.print(f"filters   : {', '.join(map(str, kwargs['filter']))}")
    else:
        kwargs["filter"] = ("",)


@click.command()
@click.argument("videos", nargs=-1, type=click.Path())
@click.option("-d", "--database", type=click.Path(), help="The database path.")
@click.option(
    "-r", "--repeat",
    type=int,
    default=5,
    help="The number of times the decoding is repeated on this machine.",
)
@click.option(
    "--resolution",
    type=ResolutionParamType(),
    multiple=True,
    help="The optional video shape conversion during decoding.",
)
@click.option(
    "--filter",
    type=str,
    multiple=True,
    help="The ffmpeg filter to apply on fly.",
)
def main(videos: tuple[str], database: str = None, **kwargs):
    """Measures activity during decoding.

    \b
    Parameters
    ----------
    videos : tuple[pathlike], optional
        All videos to be decoded. It can be a glob expression, a directory or a file path.
    database : pathlike, optional
        The path to the database where all measurements are stored.
        If a folder is provided, the database is created inside this folder.
    repeat : int, default=5
        The number of times the experiment is repeated on this environement.
        This allows us to estimate the variance of measurements.
    resolution : tuple[int, int], optional
        If provided, the decoded video will be reshaped on fly.
        If this argument is not provided, the resolution remains unchanged.
    filter : str, optional
        The additional ffmpeg video filter (after the -vf command)
        to be applied immediately after decoding and before other conversions.
    """
    with Printer("Parse configuration...") as prt:
        videos, database = parse_videos_database(prt, videos, database)
        _parse_args(prt, kwargs)

    # preparation of context
    env_id = add_environement(database)
    kwargs["dec_vid_id"]: dict[pathlib.Path, bytes] = compute_video_hash(videos)

    # retrieves the settings for videos that have already been decoded
    with sqlite3.connect(database) as sql_database:
        sql_database.row_factory = sqlite3.Row
        done: dict[tuple, int] = {}
        for values in sql_database.execute(
            "SELECT * FROM t_dec_decode WHERE dec_env_id=?", (env_id,)
        ):
            values = (
                values["dec_vid_id"],
                values["dec_filter"],
                (values["dec_height"], values["dec_width"]),
            )
            done[values] = done.get(values, 0) + 1

    # iterate on all the parameters
    loops = kwargs["repeat"] * len(videos) * len(kwargs["filter"]) * len(kwargs["resolution"])
    for i, values in enumerate(
        itertools.product(range(kwargs["repeat"]), videos, kwargs["filter"], kwargs["resolution"])
    ):
        values = dict(zip(("repeat", "video", "filter", "resolution"), values))
        values["dec_vid_id"] = kwargs["dec_vid_id"][values["video"]]
        values["filter"] = values["filter"] or ""
        values["resolution"] = values["resolution"] or get_resolution(values["video"])
        key = (values["dec_vid_id"], values["resolution"])
        if done.get(key, 0) > values["repeat"]:
            continue
        with Printer(f"Decode {i+1}/{loops}...", color="cyan") as prt:
            decode_and_store(database, env_id, values.pop("video"), **values)
            done[key] = done.get(key, 0) + 1
            prt.print_time()



#!/usr/bin/env python3

"""Generate a Python code that allows you to plot the graph."""

import ast
import datetime
import itertools
import logging
import pathlib

from .axis import get_label_extractor
from .extract import SqlLinker


def printer(**kwargs) -> str:
    """Create an excecutable python code.

    Parameters
    ----------
    database : pathlib.Path
        The database path.
    x, y : tuple[str]
        The name of the var along each axis.
    """
    # create code
    code: list[str] = []
    code.extend(print_header())
    code.extend(print_import())
    code.extend(print_cst(**kwargs))
    if kwargs["filter"] is not None:
        code.extend(print_selector(**kwargs))
    code.extend(print_read_sql(**kwargs))
    code.extend(print_fill_axe(**kwargs))
    code.extend(print_main(**kwargs))
    code.extend(print_entry())
    code = "\n".join(code)

    # format code
    try:
        import black
    except ImportError:
        logging.warning("failed to import black, please install it (uv pip install black)")
    else:
        code = black.format_str(code, mode=black.FileMode(line_length=100))
    return code


def print_entry() -> list[str]:
    """Return the code of the entry point."""
    return [
        'if __name__ == "__main__":',
        "    values: dict[str] = read_sql(DATABASE)",
        "    main(values)",
        "",
    ]


def print_cst(**kwargs) -> list[str]:
    """Return the code of the definition of the constants."""
    return [
        "# COLORS = list(mcolors.TABLEAU_COLORS.values())  # list of colors to be plotted",
        "COLORS = ['#ef597b', '#29a2c6', '#ffcb18', '#73b66b', '#4e4f4b']",
        f"DATABASE = pathlib.Path({str(kwargs['database'])!r})",
        f"FIGNAME = pathlib.Path(__file__).with_suffix('.pdf')",
        "MARKERS = ['o', 'x', '+', '^', 'v']  # list of 'dot' shapes",
        "MARKERSIZE = 4",
        "",
        "",
    ]


def print_fill_axe(**kwargs) -> list[str]:
    """Return the code that fill a given axes."""
    code = [
        f"def fill_axe(axe: plt.Axes, values: dict[str], ylab: str, xlab: str):",
        '    """Fill in the axis with the data provided as input."""',
    ]
    if kwargs["color"] is not None:
        code.append(f"    color_labels = sorted(set(values[{kwargs['color']!r}]))")
    if kwargs["marker"] is not None:
        code.append(f"    marker_labels = sorted(set(values[{kwargs['marker']!r}]))")
    match (kwargs["color"] is None, kwargs["marker"] is None):
        case (True, True):  # no color, no marker
            code.extend([
                "    x_data, y_data = values[xlab], values[ylab]",
                "    axe.errorbar(",
                "        x_data, y_data,",
                "        color=COLORS[0],",
                "        fmt=MARKERS[0],",
                "        capsize=3,",
                "        markersize=MARKERSIZE,",
                "    )",
            ])
        case (False, True):  # only color
            code.extend([
                "    for color_label, color in zip(color_labels, itertools.cycle(COLORS)):",
                f"        mask = [v == color_label for v in values[{kwargs['color']!r}]]",
                "        sub_values = {k: [v for v, m in zip(vs, mask) if m] for k, vs in values.items()}",
                "        x_data, y_data = sub_values[xlab], sub_values[ylab]",
                f"        axe.errorbar(",
                "            x_data, y_data,",
                f'            label=f"{kwargs['color']}={{color_label}}", ',
                "            color=color,",
                "            fmt=MARKERS[0],",
                "            capsize=3,",
                "            markersize=MARKERSIZE",
                "        )",
            ])
        case (True, False):  # only marker
            code.extend([
                "    for marker_label, marker in zip(marker_labels, itertools.cycle(MARKERS)):",
                f"        mask = [v == marker_label for v in values[{kwargs['marker']!r}]]",
                "        sub_values = {k: [v for v, m in zip(vs, mask) if m] for k, vs in values.items()}",
                "        x_data, y_data = sub_values[xlab], sub_values[ylab]",
                f"        axe.errorbar(",
                "            x_data, y_data,",
                f'            label=f"{kwargs['marker']}={{marker_label}}", ',
                "            color=COLORS[0],",
                "            fmt=marker,",
                "            capsize=3,",
                "            markersize=MARKERSIZE",
                "        )",
            ])
        case (False, False):  # both color and marker
            code.extend([
                "",
                "    # create legend",
                "    for color_label, color in zip(color_labels, itertools.cycle(COLORS)):",
                f"        axe.errorbar([], [], color=color, label=f'{kwargs['color']}={{color_label}}')",
                "    for marker_label, marker in zip(marker_labels, itertools.cycle(MARKERS)):",
                f"        axe.errorbar([], [], color='black', fmt=marker, label=f'{kwargs['marker']}={{marker_label}}')",
                "",
                "    # plot data",
                "    for (color_label, color), (marker_label, marker) in itertools.product(",
                "        zip(color_labels, itertools.cycle(COLORS)),",
                "        zip(marker_labels, itertools.cycle(MARKERS)),",
                "    ):",
                "        mask = [",
                "            vc == color_label and vm == marker_label",
                f"            for vc, vm in zip(values[{kwargs['color']!r}], values[{kwargs['marker']!r}])",
                "        ]",
                "        sub_values = {k: [v for v, m in zip(vs, mask) if m] for k, vs in values.items()}",
                "        x_data, y_data = sub_values[xlab], sub_values[ylab]",
                f"        axe.errorbar(",
                "            x_data, y_data,",
                "            color=color,",
                "            fmt=marker,",
                "            capsize=3,",
                "            markersize=MARKERSIZE",
                "        )",
            ])

    code.extend(["", ""])
    return code


def print_header() -> list[str]:
    """Return the code at the very top of the file."""
    return [
        "#!/usr/bin/env python3",
        "",
        f'"""Code auto-generated by mendevi the {datetime.datetime.now()}."""',
        "",
    ]


def print_import() -> list[str]:
    """Print the importations."""
    return [
        "import itertools",
        "import pathlib",
        "import sqlite3",
        "",
        "import matplotlib.colors as mcolors",
        "import matplotlib.pyplot as plt",
        "import mendevi.plot.extract as extract",
        "",
        "",
    ]


def print_selector(**kwargs) -> list[str]:
    """Return the code for selecting a subset of the data."""
    names = sorted(
        {
            n.id
            for n in ast.walk(ast.parse(kwargs["filter"], mode="eval"))
            if isinstance(n, ast.Name)
        }
    )
    if len(names) == 1:
        loop = f"    for {names[0]} in values[{names[0]!r}]:"
    else:
        loop = f"    for {', '.join(names)} in zip({', '.join(f'values[{n!r}]' for n in names)}):"
    return [
        "def select(values: dict[str]) -> dict[str]:",
        '    """Keep only some of the data."""',
        "    mask: list[bool] = []",
        loop,
        f"        mask.append(bool({kwargs['filter']}))",
        "    values = {k: [v for v, m in zip(vs, mask) if m] for k, vs in values.items()}",
        "    return values",
        "",
        "",
    ]


def print_subplot(**kwargs) -> list[str]:
    """Return the code for the subplot of a a given y and y axis."""
    header = [
        f"def plot_{kwargs['ylabel']}_{kwargs['xlabel']}(values: dict[str]):",
        f'    """Create the subchart for y={kwargs["ylabel"]} and x={kwargs["xlabel"]}."""',
    ]
    legend = [
        "        leg = fig.legend(lines, labels, loc='outside upper center', ncols=min(3, len(labels)))",
        "        leg.set_in_layout(True)",
    ]
    match (kwargs["window_y"] is None, kwargs["window_x"] is None):
        case (True, True):
            middle = [
                "    # create a simple figure",
                "    fig, axe = plt.subplots(sharex=True, figsize=(8, 8), constrained_layout=True)",
                f"    fill_axe(axe, values, ylab={kwargs['ylabel']!r}, xlab={kwargs['xlabel']!r})",
                "    lines, labels = axe.get_legend_handles_labels()",
            ]
            legend = ["        axe.legend()", "        fig.set_constrained_layout(True)"]
        case (False, True):
            middle = [
                "    # create a 1d vertical subplot figure",
                f"    all_{kwargs['window_y']}_values = sorted(set(values[{kwargs['window_y']!r}]))",
                f"    nrows = len(all_{kwargs['window_y']}_values)",
                "    fig, axes = plt.subplots(nrows=nrows, sharex=True, figsize=(8, 8), constrained_layout=True)",
                "",
                "    # iterate over all subplots",
                f"    for i, {kwargs['window_y']}_value in enumerate(all_{kwargs['window_y']}_values):",
                "        # select the correct data subset",
                f"        mask = [y == {kwargs['window_y']}_value for y in values[{kwargs['window_y']!r}]]",
                "        sub_values = {k: [v for v, m in zip(vs, mask) if m] for k, vs in values.items()}",
                f"        axes[i].set_title({kwargs['window_y']}_value)",
                f"        fill_axe(axes[i], sub_values, ylab={kwargs['ylabel']!r}, xlab={kwargs['xlabel']!r})",
                "    lines, labels = axes[0].get_legend_handles_labels()",
            ]
        case (True, False):
            middle = [
                "    # create a 1d horizontal subplot figure",
                f"    all_{kwargs['window_x']}_values = sorted(set(values[{kwargs['window_x']!r}]))",
                f"    ncols = len(all_{kwargs['window_x']}_values)",
                "    fig, axes = plt.subplots(ncols=ncols, sharey=True, figsize=(8, 8), constrained_layout=True)",
                "",
                "    # iterate over all subplots",
                f"    for j, {kwargs['window_x']}_value in enumerate(all_{kwargs['window_x']}_values):",
                "        # select the correct data subset",
                f"        mask = [x == {kwargs['window_x']}_value for x in values[{kwargs['window_x']!r}]]",
                "        sub_values = {k: [v for v, m in zip(vs, mask) if m] for k, vs in values.items()}",
                f"        axes[j].set_title({kwargs['window_x']}_value)",
                f"        fill_axe(axes[j], sub_values, ylab={kwargs['ylabel']!r}, xlab={kwargs['xlabel']!r})",
                "    lines, labels = axes[0].get_legend_handles_labels()",
            ]
        case (False, False):  # 2d grid subplot
            middle = [
                "    # create a 2d grid subplot figure",
                f"    all_{kwargs['window_y']}_values = sorted(set(values[{kwargs['window_y']!r}]))",
                f"    all_{kwargs['window_x']}_values = sorted(set(values[{kwargs['window_x']!r}]))",
                f"    nrows, ncols = len(all_{kwargs['window_y']}_values), len(all_{kwargs['window_x']}_values)",
                "    fig, axes = plt.subplots(nrows, ncols, sharex=True, sharey=True, figsize=(8, 8), constrained_layout=True)",
                "",
                "    # iterate over all subplots",
                f"    for (i, {kwargs['window_y']}_value), (j, {kwargs['window_x']}_value) in itertools.product(",
                f"        enumerate(all_{kwargs['window_y']}_values), enumerate(all_{kwargs['window_x']}_values)",
                "    ):",
                "        # select the correct data subset",
                "        mask = [",
                f"            y == {kwargs['window_y']}_value and x == {kwargs['window_x']}_value",
                f"            for y, x in zip(values[{kwargs['window_y']!r}], values[{kwargs['window_x']!r}])",
                "        ]",
                "        sub_values = {k: [v for v, m in zip(vs, mask) if m] for k, vs in values.items()}",
                f"        axes[i, j].set_title(f'{{{kwargs['window_y']}_value}} {{{kwargs['window_x']}_value}}')",
                f"        fill_axe(axes[i, j], sub_values, ylab={kwargs['ylabel']!r}, xlab={kwargs['xlabel']!r})",
                "    lines, labels = axes[0, 0].get_legend_handles_labels()",
            ]
    end = [
        "    if labels:",
        *legend,
        f"    fig.supylabel({get_label_extractor(kwargs['ylabel'])[0]!r})",
        f"    fig.supxlabel({get_label_extractor(kwargs['xlabel'])[0]!r})",
        "",
    ]
    return header + middle + end


def print_main(**kwargs) -> list[str]:
    """Return the code of the function that plot the final graphic."""
    code = []
    # subplots
    for ylabel, xlabel in itertools.product(kwargs["y"], kwargs["x"]):
        code.extend(print_subplot(**kwargs, ylabel=ylabel, xlabel=xlabel))

    # main function
    code.extend([
        "def main(values: dict[str]):",
        '    """Create the plots with matplotlib."""',
    ])
    for (i, ylabel), (j, xlabel) in itertools.product(
        enumerate(kwargs["y"]), enumerate(kwargs["x"])
    ):
        code.extend([
            "    # TODO: display in a grid rather than creating multiple windows",
            f"    plot_{ylabel}_{xlabel}(values)  # in position rows={i+1}, cols={j+1}",
        ])
    code.extend([
        "    plt.savefig(FIGNAME, format=FIGNAME.suffix[1:])",
        "    plt.show()",
        "",
        "",
    ])
    return code


def print_read_sql(**kwargs) -> list[str]:
    """Return the code of the function that perform the sql request."""
    names = set(kwargs["x"]) | set(kwargs["y"])
    if kwargs["color"] is not None:
        names.add(kwargs["color"])
    if kwargs["marker"] is not None:
        names.add(kwargs["marker"])
    if kwargs["window_x"] is not None:
        names.add(kwargs["window_x"])
    if kwargs["window_y"] is not None:
        names.add(kwargs["window_y"])
    if kwargs["filter"] is not None:
        names |= {
            n.id
            for n in ast.walk(ast.parse(kwargs["filter"], mode="eval"))
            if isinstance(n, ast.Name)
        }
    names = sorted(names)

    # get sql query
    funcs = {n: get_label_extractor(n)[1] for n in names}
    select = sorted({s for f in funcs.values() for s in f.select})
    if len(table := {f.table for f in funcs.values()}) != 1:
        logging.warning("invalid sql query because select from 2 differents tables")
        table = ", ".join(sorted(table))
    else:
        table = table.pop()
    join = sorted({j for f in funcs.values() for j in f.join})
    query = SqlLinker(select, table, join).sql

    # selector
    selector = [] if kwargs["filter"] is None else ["    values = select(values)"]

    # include in template
    extract = (f"values[{n!r}].append(extract.{funcs[n].__name__}(raw))" for n in names)
    return [
        "def read_sql(database: pathlib.Path) -> dict[str, list]:",
        '    """Extract the relevant values from the database."""',
        f"    values: dict[str, list] = {{{', '.join(f'{n!r}: []' for n in names)}}}",
        "    with sqlite3.connect(database) as conn:",
        "        conn.row_factory = sqlite3.Row",
        "        for raw in conn.execute(",
        '            """',
        f"            {'\n            '.join(query.split('\n'))}",
        '            """',
        "        ):",
        "            raw = dict(raw)",
        f"            {'\n            '.join(extract)}",
        *selector,
        "    return values",
        "",
        "",
    ]



#!/usr/bin/env python3

"""Perform encoding measures."""

import fractions
import functools
import itertools
import operator
import pathlib
import sqlite3

from context_verbose import Printer
import click
import numpy as np

from mendevi.database.complete import add_environement
from mendevi.encode import encode_and_store
from mendevi.utils import compute_video_hash, get_pix_fmt, get_rate_video, get_resolution
from .parse import PixelParamType, ResolutionParamType, parse_videos_database



ENCODERS = {"libx264", "libx265", "libvpx-vp9", "libsvtav1", "vvc"}



def _parse_args(prt: Printer, kwargs: dict):
    """Verification of the arguments."""
    # repeat
    assert "repeat" in kwargs, sorted(kwargs)
    assert isinstance(kwargs["repeat"], int), kwargs["repeat"].__class__.__name__
    assert kwargs["repeat"] >= 1, kwargs["repeat"]
    prt.print(f"repeat    : {kwargs['repeat']}")

    # effort
    assert "effort" in kwargs, sorted(kwargs)
    assert isinstance(kwargs["effort"], tuple), kwargs["effort"].__class__.__name__
    assert all(isinstance(p, str) for p in kwargs["effort"]), kwargs["effort"]
    assert all(p in {"fast", "medium", "slow"} for p in kwargs["effort"]), kwargs["effort"]
    prt.print(f"efforts   : {', '.join(kwargs['effort'])}")

    # encoder
    assert "encoder" in kwargs, sorted(kwargs)
    assert isinstance(kwargs["encoder"], tuple), kwargs["encoder"].__class__.__name__
    assert all(isinstance(e, str) for e in kwargs["encoder"]), kwargs["encoder"]
    assert all(e in ENCODERS for e in kwargs["encoder"]), kwargs["encoder"]
    prt.print(f"encoders  : {', '.join(kwargs['encoder'])}")

    # points
    assert "points" in kwargs, sorted(kwargs)
    assert isinstance(kwargs["points"], int), kwargs["points"].__class__.__name__
    assert kwargs["points"] >= 1, kwargs["points"]
    kwargs["quality"] = np.linspace(
        1.0/(kwargs["points"]+1),
        kwargs["points"]/(kwargs["points"]+1),
        kwargs["points"],
        dtype=np.float16,
    ).tolist()
    prt.print(f"qualities : k/{kwargs['points']+1}, k \u2208 [1, {kwargs['points']}]")

    # threads
    assert "threads" in kwargs, sorted(kwargs)
    assert isinstance(kwargs["threads"], tuple), kwargs["threads"].__class__.__name__
    assert all(isinstance(t, int) for t in kwargs["threads"]), kwargs["threads"]
    assert all(t >= 1 for t in kwargs["threads"]), kwargs["threads"]
    prt.print(f"threads   : {', '.join(map(str, kwargs['threads']))}")

    # fps
    assert "fps" in kwargs, sorted(kwargs)
    assert isinstance(kwargs["fps"], tuple), kwargs["fps"].__class__.__name__
    assert all(isinstance(f, fractions.Fraction) for f in kwargs["fps"]), kwargs["fps"]
    assert all(f > 0 for f in kwargs["fps"]), kwargs["fps"]
    if kwargs["fps"]:
        kwargs["fps"] = [f.limit_denominator(1001) for f in kwargs["fps"]]
        prt.print(f"fps       : {', '.join(map(str, kwargs['fps']))}")
    else:
        kwargs["fps"] = (None,)

    # resolution
    assert "resolution" in kwargs, sorted(kwargs)
    assert isinstance(kwargs["resolution"], tuple), kwargs["resolution"].__class__.__name__
    assert all(
        isinstance(r, tuple) and len(r) == 2 and all(isinstance(s, int) and s > 0 for s in r)
        for r in kwargs["resolution"]
    ), kwargs["resolution"]
    if kwargs["resolution"]:
        prt.print(f"resolution: {', '.join(f'w={w}:h={h}' for h, w in kwargs['resolution'])}")
    else:
        kwargs["resolution"] = (None,)

    # pix_fmt
    assert "pix_fmt" in kwargs, sorted(kwargs)
    assert isinstance(kwargs["pix_fmt"], tuple), kwargs["pix_fmt"].__class__.__name__
    assert all(isinstance(p, str) for p in kwargs["pix_fmt"]), kwargs["pix_fmt"]
    if kwargs["pix_fmt"]:
        prt.print(f"pix_fmt: {', '.join(kwargs['pix_fmt'])}")
    else:
        kwargs["pix_fmt"] = (None,)

    # filter
    kwargs["filter"] = kwargs.get("filter", ())
    assert isinstance(kwargs["filter"], tuple), kwargs["filter"].__class__.__name__
    assert all(isinstance(f, str) for f in kwargs["filter"]), kwargs["filter"]
    if kwargs["filter"]:
        prt.print(f"filters   : {', '.join(map(str, kwargs['filter']))}")
    else:
        kwargs["filter"] = ("",)


@click.command()
@click.argument("videos", type=click.Path(), nargs=-1)
@click.option("-d", "--database", type=click.Path(), help="The database path.")
@click.option(
    "-r", "--repeat",
    type=int,
    default=2,
    help="The number of times the encoding is repeated on this machine.",
)
@click.option(
    "-e", "--effort",
    type=click.Choice(["fast", "medium", "slow"]),
    default=["medium"],
    multiple=True,
    help="The compression effort (default = medium).",
)
@click.option(
    "-c", "--encoder",
    type=click.Choice(sorted(ENCODERS)),
    default=sorted(ENCODERS),
    multiple=True,
    help="The encoder name.",
)
@click.option(
    "-n", "--points",
    type=int,
    default=24,
    help="The number of quality point per encoder.",
)
@click.option(
    "-t", "--threads",
    type=int,
    default=[8],
    multiple=True,
    help="The number of threads used by encoders.",
)
@click.option(
    "-f", "--fps",
    type=fractions.Fraction,
    multiple=True,
    help="The optional framerate conversion during encoding.",
)
@click.option(
    "--resolution",
    type=ResolutionParamType(),
    multiple=True,
    help="The optional video shape conversion during encoding.",
)
@click.option(
    "--pix_fmt",
    type=PixelParamType(),
    multiple=True,
    help="The optional pixel format conversion during encoding.",
)
@click.option(
    "--filter",
    type=str,
    multiple=True,
    help="The ffmpeg filter to apply before encoding.",
)
def main(videos: tuple, database: str = None, **kwargs):
    """Measures activity during encoding.

    \b
    Parameters
    ----------
    videos : tuple[pathlike]
        The source videos to be transcoded.
        It can be a glob expression, a directory or a file path.
    database : pathlike, optional
        The path to the database where all measurements are stored.
        If a folder is provided, the database is created inside this folder.
        By default, it is created right next to the video.
    repeat : int, default=2
        The number of times the experiment is repeated on this environement.
        This allows us to estimate the variance of measurements.
    effort : tuple[str], default=("medium",)
        The effort made to compress, `fast`, `medium` or `slow`.
    encoder : tuple[str], default=("libsvtav1", "libvpx-vp9", "libx264", "libx265", "vvc")
        The encoders and therefore the codecs to use.
    points : int, default=24
        The number of different qualities to use.
        It is an indirect way to determine the CRF or the QP.
        The quality values are distributed evenly over ]0, 1[,
        for example, points=3 => qualities=[0.25, 0.5, 0.75].
    threads : int, default=8
        The theoretical number of threads used by the encoder.
        This roughly reflects the number of logical cores used.
    fps : fractions.Fraction, optional
        If provided, the reference video will be resampled on fly just before being encoded.
        This is therefore the frame rate of the transcoded video.
        If this argument is not provided, the frame rate remains unchanged.
    resolution : tuple[int, int], optional
        If provided, the reference video will be reshaped on fly just before being encoded.
        This is therefore the resolution of the transcoded video.
        If this argument is not provided, the resolution remains unchanged.
    pix_fmt : str, optional
        If provided, A pixel conversion is performed during encoding.
    filter : str, optional
        A video ffmpeg filter (after -vf) which applies at the time of transcoding,
        more precisely between decoding and conversion. It is a kind of pre-conversion.
    """
    with Printer("Parse configuration...") as prt:
        videos, database = parse_videos_database(prt, videos, database)
        assert videos, "at least one video is required"
        _parse_args(prt, kwargs)

    # preparation of context
    env_id = add_environement(database)
    kwargs["src_vid_id"]: dict[pathlib.Path, bytes] = compute_video_hash(videos)

    keys = ["effort", "encoder", "filter", "fps", "pix_fmt", "quality", "resolution", "threads"]

    # retrieves the settings for videos that have already been transcoded
    with sqlite3.connect(database) as sql_database, Printer("Search already done...") as prt:
        sql_database.row_factory = sqlite3.Row
        done: dict[tuple, int] = {}
        for values in sql_database.execute(
            "SELECT * FROM t_enc_encode WHERE enc_env_id=?", (env_id,)
        ):
            values = {
                "effort": values["enc_effort"],
                "encoder": values["enc_encoder"],
                "filter": values["enc_filter"] or None,
                "fps": fractions.Fraction(values["enc_fps"]).limit_denominator(1001),
                "pix_fmt": values["enc_pix_fmt"],
                "quality": float(values["enc_quality"]),
                "resolution": (values["enc_height"], values["enc_width"]),
                "src_vid_id": values["enc_src_vid_id"],
                "threads": values["enc_threads"],
            }
            values = tuple(values[k] for k in ["src_vid_id", *keys])
            done[values] = done.get(values, 0) + 1
        prt.print(f"{sum(done.values())} video already encoded under these conditions")

    # iterate on all the parameters
    loops = functools.reduce(
        operator.mul, [kwargs["repeat"], len(videos), *(len(kwargs[k]) for k in keys)]
    )
    for i, values in enumerate(
        itertools.product(range(kwargs["repeat"]), videos, *(kwargs[k] for k in keys))
    ):
        values = dict(zip(["repeat", "video", *keys], values), repeat=values[0])
        values["filter"] = values["filter"] or None
        values["fps"] = values["fps"] or get_rate_video(values["video"])
        values["pix_fmt"] = values["pix_fmt"] or get_pix_fmt(values["video"])
        values["resolution"] = values["resolution"] or get_resolution(values["video"])
        values["src_vid_id"] = kwargs["src_vid_id"][values["video"]]
        key = (values["src_vid_id"],) + tuple(values[k] for k in keys)
        if done.get(key, 0) > values["repeat"]:
            continue
        with Printer(f"Encode {i+1}/{loops}...", color="cyan") as prt:
            encode_and_store(database, env_id, values.pop("video"), **values)
            done[key] = done.get(key, 0) + 1
            prt.print_time()
